import os
import pytest
import unittest

import numpy as np
from osgeo import gdal, osr, ogr

from scipy.special import erf

from context import scarplet
import scarplet as sl
from scarplet.WindowedTemplate import Scarp, Channel


DEFAULT_EPSG = 32610 # UTM 10N
TEST_DIR = os.path.dirname(__file__)


class ScarpTestCase(unittest.TestCase):


    def setUp(self):
    
        self.obj = Scarp(100, 10, 0, 100, 100, 1)

    def test_template(self):

        test = self.obj.template()
        true = np.load(os.path.join(TEST_DIR, 'results/scarp_template.npy'))

        self.assertTrue(np.allclose(test, true), "Scarp template function is \
                        incorrect")

    @pytest.mark.xfail(reason='numexpr version differences')
    def test_template_numexpr(self):

        test = self.obj.template_numexpr()
        true = np.load(os.path.join(TEST_DIR, 'results/scarp_template_numexpr.npy'))

        self.assertTrue(np.allclose(test, true), "Scarp template function is \
                        incorrect")


class ChannelTestCase(unittest.TestCase):


    def setUp(self):
    
        self.obj = Channel(100, 0.1, 0, 100, 100, 1)

    def test_template(self):

        test = self.obj.template()
        true = np.load(os.path.join(TEST_DIR, 'results/channel_template.npy'))

        self.assertTrue(np.allclose(test, true), "Channel template function is \
                        incorrect")



# -*- coding: utf-8
""" Functions for determining best-fit template parameters by convolution with
a grid """

import numexpr
import numpy as np
import multiprocessing as mp

import matplotlib
import matplotlib.pyplot as plt

import pyfftw
from pyfftw.interfaces.numpy_fft import fft2, ifft2, fftshift

from functools import partial

from scarplet import WindowedTemplate
from scarplet.dem import DEMGrid


np.seterr(divide='ignore', invalid='ignore')

pyfftw.interfaces.cache.enable()


def calculate_amplitude(dem, Template, scale, age, angle):
    """Calculate amplitude and SNR of features using a template

    Parameters
    ----------
    dem : DEMGrid
        Grid object of elevation data
    Template : WindowedTemplate
        Class representing template function
    scale : float
        Scale of template function in DEM cell units
    age : float
        Age parameter for template function
    angle : float
        Orientation of template in radians

    Returns
    -------
    amp : np.array
        2-D array of amplitudes for each DEM pixel
    snr : np.array
        2-D array of signal-to-noise ratios for each DEM pixel
    """

    ny, nx = dem._griddata.shape
    de = dem._georef_info.dx
    t = Template(scale, age, angle, nx, ny, de)
    template = t.template()

    curv = dem._calculate_directional_laplacian(angle)

    amp, age, angle, snr = match_template(curv, template)
    mask = t.get_window_limits()
    amp[mask] = 0
    snr[mask] = 0

    return amp, snr


def calculate_best_fit_parameters_serial(dem,
                                         Template,
                                         scale,
                                         ang_max=np.pi / 2,
                                         ang_min=-np.pi / 2,
                                         **kwargs):
    """Calculate best-fitting parameters using a template

    Parameters
    ----------
    dem : DEMGrid
        Grid object of elevation data
    Template : WindowedTemplate
        Class representing template function
    scale : float
        Scale of template function in DEM cell units

    Other Parameters
    ----------------
    ang_max : float, optional
        Maximum orietnation of template, default pi / 2
    ang_min : float, optional
        Minimum orietnation of template, default -pi / 2
    kwargs : optional
        Any additional keyword arguments that may be passed to the template()
        method of the Template class

    Returns
    -------
    best_amp : np.array
        2-D array of best-fitting amplitudes for each DEM pixel
    best_age : np.array
        2-D array of best-fitting agees for each DEM pixel
    best_angle : np.array
        2-D array of best-fitting orientations for each DEM pixel
    best_snr : np.array
        2-D array of maximum signal-to-noise ratios for each DEM pixel
    """

    ang_stepsize = 1
    num_angles = int((180 / np.pi) * (ang_max - ang_min) / ang_stepsize + 1)
    orientations = np.linspace(ang_min, ang_max, num_angles)
    ages = 10 ** np.arange(0, 3.5, 0.1)

    ny, nx = dem._griddata.shape
    best_amp = np.zeros((ny, nx))
    best_angle = np.zeros((ny, nx))
    best_age = np.zeros((ny, nx))
    best_snr = np.zeros((ny, nx))

    for this_angle in orientations:
        for this_age in ages:
            this_amp, this_age, this_angle, this_snr = match_template(dem,
                                                                      Template,
                                                                      scale,
                                                                      this_age,
                                                                      this_angle,
                                                                      **kwargs)

            best_amp = numexpr.evaluate("(best_snr > this_snr)*best_amp + \
                                        (best_snr < this_snr)*this_amp")

            best_angle = numexpr.evaluate("(best_snr > this_snr)*best_angle + \
                                          (best_snr < this_snr)*this_angle")

            best_age = numexpr.evaluate("(best_snr > this_snr)*best_age + \
                                        (best_snr < this_snr)*this_age")

            best_snr = numexpr.evaluate("(best_snr > this_snr)*best_snr + \
                                        (best_snr < this_snr)*this_snr")

    return best_amp, best_age, best_angle, best_snr


def calculate_best_fit_parameters(dem,
                                  Template,
                                  scale,
                                  age,
                                  ang_max=np.pi / 2,
                                  ang_min=-np.pi / 2,
                                  **kwargs):
    """Calculate best-fitting parameters using a template with parallel search

    Parameters
    ----------
    dem : DEMGrid
        Grid object of elevation data
    Template : WindowedTemplate
        Class representing template function
    scale : float
        Scale of template function in DEM cell units
    age : float
        Age parameter for template function

    Other Parameters
    ----------------
    ang_max : float, optional
        Maximum orietnation of template, default pi / 2
    ang_min : float, optional
        Minimum orietnation of template, default -pi / 2

    Returns
    -------
    results : np.array
        Array of best amplitudes, ages, orientations, and  signal-to-noise
        ratios for each DEM pixel. Dimensions of (4, height, width).
    """

    ang_stepsize = 1
    num_angles = int((180 / np.pi) * (ang_max - ang_min) / ang_stepsize + 1)
    orientations = np.linspace(ang_min, ang_max, num_angles)
    orientations = (angle for angle in orientations)

    ny, nx = dem._griddata.shape

    nprocs = mp.cpu_count()
    pool = mp.Pool(processes=nprocs)
    wrapper = partial(match_template, dem, Template, scale, age)
    results = pool.imap(wrapper, orientations, chunksize=1)

    best_amp, best_age, best_angle, best_snr = compare(results, ny, nx)

    pool.close()
    pool.join()

    results = np.stack([best_amp,
                        best_age,
                        best_angle,
                        best_snr])

    return results


def compare(results, ny, nx):
    """Compare template matching results from asynchronous tasks

    Parameters
    ----------
    results : iterable
        Iterable containing outputs of a template matching method
    ny : int
        Number of rows in output
    nx : int
        Number of columns in output

    Returns
    -------
    best_amp : np.array
        2-D array of best-fitting amplitudes
    best_age : np.array
        2-D array of best-fitting morphologic ages 
    best_angle : np.array
        2-D array of best-fitting orientations
    best_snr : np.array
        2-D array of maximum signal-to-noise ratios
    """

    best_amp = np.zeros((ny, nx))
    best_age = np.zeros((ny, nx))
    best_angle = np.zeros((ny, nx))
    best_snr = np.zeros((ny, nx))

    for r in results:
        this_amp, this_age, this_angle, this_snr = r

        best_amp = numexpr.evaluate("(best_snr > this_snr)*best_amp + \
                                    (best_snr < this_snr)*this_amp")

        best_age = numexpr.evaluate("(best_snr > this_snr)*best_age + \
                                    (best_snr < this_snr)*this_age")

        best_angle = numexpr.evaluate("(best_snr > this_snr)*best_angle + \
                                      (best_snr < this_snr)*this_angle")

        best_snr = numexpr.evaluate("(best_snr > this_snr)*best_snr + \
                                    (best_snr < this_snr)*this_snr")
        del this_amp, this_snr, r

    return best_amp, best_age, best_angle, best_snr


def load(filename):
    """Load DEM from file

    Parameters
    ----------
    filename : string
        Filename of DEM

    Returns
    -------
    data_obj : DEMGrid
        DEMGrid object with DEM data
    """

    data_obj = DEMGrid(filename)
    data_obj._fill_nodata()

    return data_obj


def match(data, Template, **kwargs):
    """Match template to input data from DEM

    Parameters
    ----------
    data : DEMGrid
        DEMGrid object containing input data
    Template : WindowedTemplate
        Class of template function to use

    Returns
    -------
    results : np.array
        Array of best amplitudes, ages, orientations, and  signal-to-noise
        ratios for each DEM pixel. Dimensions of (4, height, width).
    """
    
    if 'age' in kwargs:
        results = calculate_best_fit_parameters(data, Template, **kwargs)
    else:
        results = calculate_best_fit_parameters_serial(data, Template, **kwargs)

    return results


def match_template(data, Template, scale, age, angle, **kwargs):
    """Match template function to curvature using convolution

    Parameters
    ----------
    data : DEMGrid
        Grid object of elevation data
    Template : WindowedTemplate
        Class representing template function
    scale : float
        Scale of template function in DEM cell units
    age : float
        Age parameter for template function
    angle : float
        Orientation of template in radians

    Other Parameters
    ----------------
    kwargs : optional
        Any additional keyword arguments that may be passed to the template()
        method of the Template class

    Returns
    -------
    amp : np.array
        2-D array of amplitudes for each DEM pixel
    age : np.array
        template age in m2
    angle : np.array 
        template orientation in radians 
    snr : np.array
        2-D array of signal-to-noise ratios for each DEM pixel

    References
    ----------
    Modifies method described in 

    Hilley, G.E., DeLong, S., Prentice, C., Blisniuk, K. and Arrowsmith, 
    J.R., 2010. Morphologic dating of fault scarps using airborne 
    laser swath mapping (ALSM) data. Geophysical Research Letters, 37(4).
    https://dx.doi.org/10.1029/2009GL042044
    """

    eps = np.spacing(1)
    curv = data._calculate_directional_laplacian(angle)
    ny, nx = curv.shape
    de = data._georef_info.dx

    template_obj = Template(scale, age, angle, nx, ny, de, **kwargs)
    template = template_obj.template()

    M = numexpr.evaluate("template != 0")
    fm2 = fft2(M)
    n = np.sum(M) + eps
    del M

    fc = fft2(curv)
    ft = fft2(template)
    fc2 = fft2(numexpr.evaluate("curv**2"))
    template_sum = np.sum(numexpr.evaluate("template**2"))
    del curv, template

    xcorr = np.real(fftshift(ifft2(numexpr.evaluate("ft*fc"))))
    amp = numexpr.evaluate("xcorr/template_sum")

    T1 = numexpr.evaluate("template_sum*(amp**2)")
    T3 = fftshift(ifft2(numexpr.evaluate("fc2*fm2")))

    # XXX: Epsilon factor is added to avoid small-magnitude dvision
    error = (1/n)*numexpr.evaluate("real(T1 - 2*amp*xcorr + T3)") + eps
    snr = numexpr.evaluate("abs(T1/error)")

    if hasattr(template_obj, 'get_err_mask'):
        mask = template_obj.get_err_mask()
        snr[mask] = 0

    mask = template_obj.get_window_limits()
    amp[mask] = 0
    snr[mask] = 0

    return amp, age, angle, snr


def plot_results(data, results, az=315, elev=45, figsize=(4, 16)):
    """Plots maps of results from template matching

    Parameters
    ----------
    data : DEMGrid
        DEMGrid object containing input data
    results : np.array
        Array of best-fitting results from compare() or similar function

    Optional Parameters
    -------------------
    az : float
        Azimuth of light source for hillshade
    elev : float
        Elevation angle of light source for hillshade
    figsize : tuple
        Figure size
    """

    fig, ax = plt.subplots(2, 2, figsize=figsize)
    ax = ax.ravel()

    ls = matplotlib.colors.LightSource(azdeg=az, altdeg=elev)
    hillshade = ls.hillshade(data._griddata,
                             vert_exag=1,
                             dx=data._georef_info.dx,
                             dy=data._georef_info.dy)

    labels = ['Amplitude [m]', 'Relative age [m$^2$]',
              'Orientation [deg.]', 'Signal-to-noise ratio']
    cmaps = ['Reds', 'viridis', 'RdBu_r', 'Reds']
    for i, val in enumerate(zip(ax, labels, cmaps)):
        axis, label, cmap = val
        axis.imshow(hillshade, alpha=1, cmap='gray')
        im = axis.imshow(results[i], alpha=0.5, cmap=cmap)
        cb = plt.colorbar(im, ax=axis, shrink=0.5,
                          orientation='horizontal', label=label)
        ticks = matplotlib.ticker.MaxNLocator(nbins=3)
        cb.locater = ticks
        cb.update_ticks()



""" Convenience functions to load example datasets """

import os
import scarplet as sl


EXAMPLE_DIRECTORY = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 'data/')


def load_carrizo():
    """
    Load sample dataset containing fault scarps along the San Andreas Fault
    from the Wallace Creek section on the Carrizo Plain, California, USA

    Data downloaded from OpenTopography and collected by the B4 Lidar Project:
    https://catalog.data.gov/dataset/b4-project-southern-san-andreas-and-san-jacinto-faults
    """

    path = os.path.join(EXAMPLE_DIRECTORY, 'carrizo.tif')
    data = sl.load(path)
    return data


def load_grandcanyon():
    """
    Load sample dataset containing part of channel network in the Grand Canyon
    Arizona, USA

    Data downloaded from the Terrain Tile dataset, part of Amazon Earth on AWS
    https://registry.opendata.aws/terrain-tiles/
    """

    path = os.path.join(EXAMPLE_DIRECTORY, 'grandcanyon.tif')
    data = sl.load(path)
    return data


def load_synthetic():
    """
    Load sample dataset of synthetic fault scarp of morphologic age 10 m2 
    """

    path = os.path.join(EXAMPLE_DIRECTORY, 'synthetic.tif')
    data = sl.load(path)
    return data



import filecmp
import numpy as np
import os
import sys
import pytest
import unittest

from osgeo import gdal, osr
from scipy.special import erf

from context import scarplet
import scarplet as sl
from scarplet import dem
from scarplet.WindowedTemplate import Scarp


DEFAULT_EPSG = 32610
TEST_DIR = os.path.dirname(__file__)


class TemplateMatchingTestCase(unittest.TestCase):
    
    
    def setUp(self):
        
        self.data = sl.load(os.path.join(TEST_DIR, 'data/synthetic.tif'))

    @pytest.mark.slow
    def test_match(self):
        
        template_args = {'scale': 100,
                        'ang_max': np.pi / 2,
                        'ang_min': -np.pi / 2
                        }

        res = sl.match(self.data, Scarp, **template_args)
        amp, age, alpha, snr = res

        true = np.load(os.path.join(TEST_DIR, 'results/synthetic_match1.npy'))
        true_amp, true_age, true_alpha, true_snr = true
        
        self.assertTrue(np.allclose(amp, true_amp), "Amplitudes incorrect")
        self.assertTrue(np.allclose(age, true_age), "Ages incorrect")
        self.assertTrue(np.allclose(alpha, true_alpha), "Orientations incorrect")
        self.assertTrue(np.allclose(snr, true_snr), "SNRs incorrect")

    def test_match_single_age(self):
        
        template_args = {'scale': 100,
                         'age': 10,
                        'ang_max': np.pi / 2,
                        'ang_min': -np.pi / 2
                        }

        res = sl.match(self.data, Scarp, **template_args)
        amp, age, alpha, snr = res
        
        true = np.load(os.path.join(TEST_DIR, 'results/synthetic_match2.npy'))
        true_amp, true_age, true_alpha, true_snr = true
        
        self.assertTrue(np.allclose(amp, true_amp), "Amplitudes incorrect")
        self.assertTrue(np.allclose(age, true_age), "Ages incorrect")
        self.assertTrue(np.allclose(alpha, true_alpha), "Orientations incorrect")
        self.assertTrue(np.allclose(snr, true_snr), "SNRs incorrect")

    def test_match_template(self):
        
        template_args = {'scale': 100,
                         'age': 10,
                         'angle': 0
                        }

        res = sl.match_template(self.data, Scarp, **template_args)
        amp, age, alpha, snr = res

        true = np.load(os.path.join(TEST_DIR, 'results/synthetic_match3.npy'))
        true_amp, true_age, true_alpha, true_snr = true
        
        self.assertTrue(np.allclose(amp, true_amp), "Amplitudes incorrect")
        self.assertTrue(np.allclose(age, true_age), "Ages incorrect")
        self.assertTrue(np.allclose(alpha, true_alpha), "Orientations incorrect")
        self.assertTrue(np.allclose(snr, true_snr), "SNRs incorrect")


def generate_synthetic_scarp(a, b, kt, x_max, y_max, de=1, sig2=0, theta=0):
    """ Generate DEM of synthetic scarp for testing """
    
    nx = int(2 * x_max / de)
    ny = int(2 * y_max / de)
    x = np.linspace(-x_max, x_max, num=nx)
    y = np.linspace(-y_max, y_max, num=ny)
    x, y = np.meshgrid(x, y)
    
    theta = np.pi/2 - theta
    xrot = x*np.cos(theta) + y*np.sin(theta)
    yrot = -x*np.sin(theta) + y*np.cos(theta)

    z = -erf(yrot/(2*np.sqrt(kt))) + b*yrot
    z = z + sig2*np.random.randn(ny, nx)

    return set_up_grid(z, nx, ny, de) 


def set_up_grid(data, nx, ny, de):

    synthetic = dem.DEMGrid()
    geo_transform = (0, de, 0, 0, 0, -de) 
    projection = osr.SpatialReference()
    projection.ImportFromEPSG(DEFAULT_EPSG)

    synthetic._griddata = data 
    synthetic._georef_info.geo_transform = geo_transform
    synthetic._georef_info.projection = projection
    synthetic._georef_info.dx = de 
    synthetic._georef_info.dy = de 
    synthetic._georef_info.nx = nx 
    synthetic._georef_info.ny = ny
    synthetic._georef_info.xllcenter = 0 
    synthetic._georef_info.yllcenter = 0 

    return synthetic



# -*- coding: utf-8
""" Utility classes and funcitons for template matching framework. """


class BoundingBox(object):

    def __init__(self, lr, ul):

        self.lrx = lr[0]
        self.lry = lr[1]
        self.lr = lr

        self.ulx = ul[0]
        self.uly = ul[1]
        self.ul = ul

        self.llx = ul[0]
        self.lly = lr[1]
        self.ll = (self.llx, self.lly)

        self.urx = lr[0]
        self.ury = ul[1]
        self.ur = (self.urx, self.ury)

        self.corners = [self.ul, self.ll, self.ur, self.lr]

    def contains(self, point):

        intersect_x = point[0] >= self.ulx and point[0] <= self.lrx
        intersect_y = point[1] >= self.lry and point[1] <= self.uly

        return intersect_x and intersect_y

    def intersects(self, bbox):

        for corner in bbox.corners:
            if self.contains(corner):
                return True

        for corner in self.corners:
            if bbox.contains(corner):
                return True

        return False



import os
import sys
import filecmp
import numpy as np
import unittest

import matplotlib
import matplotlib.pyplot as plt

from context import scarplet
from scarplet import dem


TEST_DIR = os.path.dirname(__file__)


class CalculationMethodsTestCase(unittest.TestCase):
    
    
    def setUp(self):
        
        self.dem = dem.DEMGrid(os.path.join(TEST_DIR, 'data/faultzone.tif'))

    def test_calculate_slope(self):

        sx, sy = self.dem._calculate_slope()
        true_sx, true_sy = np.load(os.path.join(TEST_DIR, 'results/faultzone_sxsy.npy'))

        self.assertTrue(np.allclose(sx, true_sx), "Slope (x direction) incorrect")
        self.assertTrue(np.allclose(sy, true_sy), "Slope (y direction) incorrect")

    def test_calculate_laplacian(self):

        del2z = self.dem._calculate_laplacian()
        true_del2z = np.load(os.path.join(TEST_DIR, 'results/faultzone_del2z.npy'))
        self.assertTrue(np.allclose(del2z, true_del2z), "Laplacian incorrect (y axis direction)")
    
    def test_calculate_directional_laplacian(self):
        
        alphas = [-np.pi / 2, -np.pi / 4, np.pi / 4, np.pi / 2]
        for alpha in alphas:
            del2z = self.dem._calculate_directional_laplacian(alpha)
            alpha *= 180 / np.pi
            true_del2z = np.load(os.path.join(TEST_DIR, 'results/faultzone_del2z_{:.0f}.npy'.format(alpha)))
            self.assertTrue(np.allclose(del2z, true_del2z), "Laplacian incorrect (+{:.0f} deg)".format(alpha))

    def test_pad_boundary(self):
        
        dx = 5
        dy = 5
        grid = self.dem._griddata

        padded_grid = np.pad(grid, pad_width=(dy, dx), mode='reflect')
        
        self.dem._pad_boundary(dx, dy)
        
        self.assertEqual(self.dem._griddata.all(), padded_grid.all(), "Grid padded incorrectly")



# -*- coding: utf-8
""" Classes for loading digital elevation models as numeric grids """

import numexpr
import numpy as np
import os
import subprocess
import sys

import matplotlib
import matplotlib.pyplot as plt

from copy import copy
from osgeo import gdal, gdalconst

from rasterio.fill import fillnodata

from scarplet.utils import BoundingBox


sys.setrecursionlimit(10000)

FLOAT32_MIN = np.finfo(np.float32).min
GDAL_DRIVER_NAME = 'GTiff'


class CalculationMixin(object):
    """Mix-in class for grid calculations"""

    def __init__(self):

        pass

    def _calculate_slope(self):
        """Calculate gradient of grid in x and y directions.

        Pads boundary so as to return slope grids of same size as object's
        grid data

        Returns
        -------
            slope_x : numpy array
                slope in x direction
            slope_y : numpy array
                slope in y direction
        """

        dx = self._georef_info.dx
        dy = self._georef_info.dy

        PAD_DX = 2
        PAD_DY = 2

        self._pad_boundary(PAD_DX, PAD_DY)
        z_pad = self._griddata

        slope_x = (z_pad[1:-1, 2:] - z_pad[1:-1, :-2]) / (2 * dx)
        slope_y = (z_pad[2:, 1:-1] - z_pad[:-2, 1:-1]) / (2 * dy)

        return slope_x, slope_y

    def _calculate_laplacian(self):
        """Calculate curvature of grid in y direction.
        """

        return self._calculate_directional_laplacian(0)

    def _calculate_directional_laplacian(self, alpha):
        """Calculate curvature of grid in arbitrary direction.

        Parameters
        ----------
            alpha : float
                direction angle (azimuth) in radians. 0 is north or y-axis.

        Returns
        -------
            del2s : numpy array
                grid of curvature values
        """

        dx = self._georef_info.dx
        dy = self._georef_info.dy
        z = self._griddata
        nan_idx = np.isnan(z)
        z[nan_idx] = 0

        dz_dx = np.diff(z, 1, 1)/dx
        d2z_dxdy = np.diff(dz_dx, 1, 0)/dx
        pad_x = np.zeros((d2z_dxdy.shape[0], 1))
        d2z_dxdy = np.hstack([pad_x, d2z_dxdy])
        pad_y = np.zeros((1, d2z_dxdy.shape[1]))
        d2z_dxdy = np.vstack([pad_y, d2z_dxdy])

        d2z_dx2 = np.diff(z, 2, 1)/dx**2
        pad_x = np.zeros((d2z_dx2.shape[0], 1))
        d2z_dx2 = np.hstack([pad_x, d2z_dx2, pad_x])

        d2z_dy2 = np.diff(z, 2, 0)/dy**2
        pad_y = np.zeros((1, d2z_dy2.shape[1]))
        d2z_dy2 = np.vstack([pad_y, d2z_dy2, pad_y])

        del2z = d2z_dx2 * np.cos(alpha) ** 2 - 2 * d2z_dxdy * np.sin(alpha) \
            * np.cos(alpha) + d2z_dy2 * np.sin(alpha) ** 2
        del2z[nan_idx] = np.nan

        return del2z

    def _calculate_directional_laplacian_numexpr(self, alpha):
        """Calculate curvature of grid in arbitrary direction.

        Optimized with numexpr expressions.

        Parameters
        ----------
            alpha : float
                direction angle (azimuth) in radians. 0 is north or y-axis.

        Returns
        -------
            del2s : numpyarray
                grid of curvature values
        """

        dx = self._georef_info.dx
        dy = self._georef_info.dy
        z = self._griddata
        nan_idx = np.isnan(z)
        z[nan_idx] = 0

        dz_dx = np.diff(z, 1, 1)/dx
        d2z_dxdy = np.diff(dz_dx, 1, 0)/dx
        pad_x = np.zeros((d2z_dxdy.shape[0], 1))
        d2z_dxdy = np.hstack([pad_x, d2z_dxdy])
        pad_y = np.zeros((1, d2z_dxdy.shape[1]))
        d2z_dxdy = np.vstack([pad_y, d2z_dxdy])

        d2z_dx2 = np.diff(z, 2, 1)/dx**2
        pad_x = np.zeros((d2z_dx2.shape[0], 1))
        d2z_dx2 = np.hstack([pad_x, d2z_dx2, pad_x])

        d2z_dy2 = np.diff(z, 2, 0)/dy**2
        pad_y = np.zeros((1, d2z_dy2.shape[1]))
        d2z_dy2 = np.vstack([pad_y, d2z_dy2, pad_y])

        del2z = numexpr.evaluate("d2z_dx2*cos(alpha)**2 - \
                2*d2z_dxdy*sin(alpha)*cos(alpha) + d2z_dy2*sin(alpha)**2")
        del2z[nan_idx] = np.nan

        return del2z

    def _estimate_curvature_noiselevel(self):
        """Estimate noise level in curvature of grid as a function of direction.

        Returns
        -------
            angles : numpy array
                array of orientations (azimuths) in radians
            mean : float
                array of mean curvature in correponding direction
            sd : float
                array of curvature standard deviation in correponding direction
        """

        from scipy import ndimage

        angles = np.linspace(0, np.pi, num=180)

        mean = []
        sd = []

        for alpha in angles:
            del2z = self._calculate_directional_laplacian(alpha)
            lowpass = ndimage.gaussian_filter(del2z, 100)
            highpass = del2z - lowpass
            mean.append(np.nanmean(highpass))
            sd.append(np.nanstd(highpass))

        return angles, mean, sd

    def _pad_boundary(self, dx, dy):
        """Pad grid boundary with reflected boundary conditions.
        """

        self._griddata = np.pad(self._griddata, pad_width=(dy, dx),
                                mode='reflect')
        self.padded = True
        self.pad_dx = dx
        self.pad_dy = dy

        ny, nx = self._griddata.shape

        self._georef_info.nx = nx
        self._georef_info.ny = ny
        self._georef_info.xllcenter -= dx
        self._georef_info.yllcenter -= dy


class GDALMixin(object):
    pass


class GeorefInfo(object):

    def __init__(self):

        self.geo_transform = None
        self.projection = None
        self.xllcenter = None
        self.yllcenter = None
        self.dx = None
        self.dy = None
        self.nx = None
        self.ny = None
        self.ulx = None
        self.uly = None
        self.lrx = None
        self.lry = None


class BaseSpatialGrid(GDALMixin):
    """Base class for spatial grid"""

    dtype = gdalconst.GDT_Float32

    def __init__(self, filename=None):

        _georef_info = GeorefInfo()

        if filename is not None:
            self._georef_info = _georef_info
            self.load(filename)
            self.filename = filename
        else:
            self.filename = None
            self._georef_info = _georef_info
            self._griddata = np.empty((0, 0))

    def is_contiguous(self, grid):
        """Returns true if grids are contiguous or overlap

        Parameters
        ----------
            grid : BaseSpatialGrid
        """

        return self.bbox.intersects(grid.bbox)

    def merge(self, grid):
        """Merge this grid with another BaseSpatialGrid.

        Wrapper argound gdal_merge.py.

        Parameters
        ----------
            grid : BaseSpatialGrid

        Returns
        -------
            merged_grid : BaseSpatialGrid
        """

        if not self.is_contiguous(grid):
            raise ValueError("ValueError: Grids are not contiguous")

        # XXX: this is hacky, eventually implement as native GDAL
        try:
            command = ['gdal_merge.py', self.filename, grid.filename]
            subprocess.check_output(command, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            print("Failed to merge grids. gdal_merge may not be installed.")
            raise e

        merged_grid = BaseSpatialGrid('out.tif')
        merged_grid._griddata[merged_grid._griddata == FLOAT32_MIN] = np.nan
        os.remove('out.tif')

        return merged_grid

    def plot(self, **kwargs):
        """Plot grid data

        Keyword args:
            Any valid keyword argument for matplotlib.pyplot.imshow
        """

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.imshow(self._griddata, **kwargs)

    def save(self, filename):
        """Save grid as georeferenced TIFF
        """

        ncols = self._georef_info.nx
        nrows = self._georef_info.ny

        driver = gdal.GetDriverByName(GDAL_DRIVER_NAME)
        out_raster = driver.Create(filename, ncols, nrows, 1, self.dtype)
        out_raster.SetGeoTransform(self._georef_info.geo_transform)
        out_band = out_raster.GetRasterBand(1)
        out_band.WriteArray(self._griddata)

        proj = self._georef_info.projection.ExportToWkt()
        out_raster.SetProjection(proj)
        out_band.FlushCache()

    def load(self, filename):
        """Load grid from file
        """

        self.label = filename.split('/')[-1].split('.')[0]

        gdal_dataset = gdal.Open(filename)
        band = gdal_dataset.GetRasterBand(1)
        nodata = band.GetNoDataValue()
        self._griddata = band.ReadAsArray().astype(float)

        if nodata is not None:
            nodata_index = np.where(self._griddata == nodata)
            if self.dtype is not np.uint8:
                self._griddata[nodata_index] = np.nan

        geo_transform = gdal_dataset.GetGeoTransform()
        projection = gdal_dataset.GetProjection()
        nx = gdal_dataset.RasterXSize
        ny = gdal_dataset.RasterYSize

        self._georef_info.geo_transform = geo_transform
        self._georef_info.projection = projection
        self._georef_info.dx = self._georef_info.geo_transform[1]
        self._georef_info.dy = self._georef_info.geo_transform[5]
        self._georef_info.nx = nx
        self._georef_info.ny = ny
        self._georef_info.xllcenter = self._georef_info.geo_transform[0] \
            + self._georef_info.dx
        self._georef_info.yllcenter = self._georef_info.geo_transform[3] \
            - (self._georef_info.ny+1) \
            * np.abs(self._georef_info.dy)

        self._georef_info.ulx = self._georef_info.geo_transform[0]
        self._georef_info.uly = self._georef_info.geo_transform[3]
        self._georef_info.lrx = self._georef_info.geo_transform[0] \
            + self._georef_info.dx * self._georef_info.nx
        self._georef_info.lry = self._georef_info.geo_transform[3] \
            + self._georef_info.dy * self._georef_info.ny
        self.bbox = BoundingBox((self._georef_info.lrx, self._georef_info.lry),
                                (self._georef_info.ulx, self._georef_info.uly))


class DEMGrid(CalculationMixin, BaseSpatialGrid):
    """Class representing grid of elevation values"""

    def __init__(self, filename=None):

        _georef_info = GeorefInfo()

        if filename is not None:
            self._georef_info = _georef_info
            self.load(filename)
            self._griddata[self._griddata == FLOAT32_MIN] = np.nan
            self.nodata_value = np.nan
            self.filename = filename
            self.shape = self._griddata.shape
            self.is_interpolated = False
        else:
            self.filename = None
            self.label = ''
            self.shape = (0, 0)
            self._georef_info = _georef_info
            self._griddata = np.empty((0, 0))
            self.is_interpolated = False

    def plot(self, color=True, **kwargs):
        fig, ax = plt.subplots(1, 1, **kwargs)

        hs = Hillshade(self)
        hs.plot()

        if color:
            im = ax.imshow(self._griddata, alpha=0.75, cmap='terrain')
            plt.colorbar(im, ax=ax, shrink=0.75, label='Elevation')

        ax.tick_params(direction='in')
        ax.set_xlabel('x')
        ax.set_ylabel('y')

    def _fill_nodata(self):
        """Fill nodata values in elevation grid by interpolation.

        Wrapper around GDAL/rasterio's FillNoData, fillnodata methods
        """

        if ~np.isnan(self.nodata_value):
            nodata_mask = self._griddata == self.nodata_value
        else:
            nodata_mask = np.isnan(self._griddata)
        self.nodata_mask = nodata_mask

        # XXX: GDAL (or rasterio) FillNoData takes mask with 0s at nodata
        num_nodata = np.sum(nodata_mask)
        prev_nodata = np.nan
        while num_nodata > 0 or num_nodata == prev_nodata:
            mask = np.isnan(self._griddata)
            col_nodata = np.sum(mask, axis=0).max()
            row_nodata = np.sum(mask, axis=1).max()
            dist = max(row_nodata, col_nodata) / 2
            self._griddata = fillnodata(self._griddata,
                                        mask=~mask,
                                        max_search_distance=dist)
            prev_nodata = copy(num_nodata)
            num_nodata = np.sum(np.isnan(self._griddata))

        self.is_interpolated = True

    def _fill_nodata_with_edge_values(self):
        """Fill nodata values using swath edge values by row,"""

        if ~np.isnan(self.nodata_value):
            nodata_mask = self._griddata == self.nodata_value
        else:
            nodata_mask = np.isnan(self._griddata)
        self.nodata_mask = nodata_mask

        for row in self._griddata:
            idx = np.where(np.isnan(row)).min()
            fill_value = row[idx]
            row[np.isnan(row)] = fill_value

        self.is_interpolated = True


class Hillshade(BaseSpatialGrid):
    """Class representing hillshade of DEM"""

    def __init__(self, dem):
        """Load DEMGrid object as Hillshade
        """

        self._georef_info = dem._georef_info
        self._griddata = dem._griddata
        self._hillshade = None

    def plot(self, az=315, elev=45):
        """Plot hillshade

        Paramaters
        ----------
            az : float
                azimuth of light source in degrees
            elev : float
                elevation angle of light source in degrees
        """

        ax = plt.gca()
        ls = matplotlib.colors.LightSource(azdeg=az, altdeg=elev)
        self._hillshade = ls.hillshade(self._griddata, vert_exag=1,
                                       dx=self._georef_info.dx,
                                       dy=self._georef_info.dy)
        ax.imshow(self._hillshade, alpha=1, cmap='gray', origin='lower')



# -*- coding: utf-8
""" Class for windowed template matching over a spatial grid """

import numexpr
import numpy as np

from scipy.special import erf, erfinv

np.seterr(divide='ignore', invalid='ignore')


class WindowedTemplate(object):
    """Base class for windowed template function

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    c : float
        Curvature limit of template
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units

    Methods
    -------
    get_coordinates():
        Get arrays of coordinates for template grid points
    get_mask():
        Get mask array giving curvature extent of template window
    get_window_limits():
        Get mask array giving window extent
    """

    def __init__(self):

        self.d = None
        self.alpha = None
        self.nx = None
        self.ny = None
        self.c = self.nx / 2. 
        self.de = None

    def get_coordinates(self):
        x = self.de * np.linspace(1, self.nx, num=self.nx)
        y = self.de * np.linspace(1, self.ny, num=self.ny)
        x = x - np.mean(x)
        y = y - np.mean(y)

        x, y = np.meshgrid(x, y)
        xr = x * np.cos(self.alpha) + y * np.sin(self.alpha)
        yr = -x * np.sin(self.alpha) + y * np.cos(self.alpha)

        return xr, yr

    def get_mask(self):
        xr, yr = self.get_coordinates()
        mask = (abs(xr) < self.c) & (abs(yr) < self.d)
        return mask

    def get_window_limits(self):

        x4 = self.d*np.cos(self.alpha - np.pi/2)
        y4 = self.d*np.sin(self.alpha - np.pi/2)
        x1 = self.d*np.cos(self.alpha)
        y1 = self.d*np.sin(self.alpha)
        an_y = abs((x4 - x1) + 2 * self.c * np.cos(self.alpha - np.pi/2))
        an_x = abs((y1 - y4) + 2 * self.c * np.sin(self.alpha - np.pi/2))

        x = self.de*np.linspace(1, self.nx, num=self.nx)
        y = self.de*np.linspace(1, self.ny, num=self.ny)
        x = x - np.mean(x)
        y = y - np.mean(y)

        X, Y = np.meshgrid(x, y)
        mask = ((X < (min(x) + an_x)) | (X > (max(x) - an_x))
                | (Y < (min(y) + an_y)) | (Y > (max(y) - an_y)))

        return mask


class Scarp(WindowedTemplate):
    """Curvature template for vertical scarp

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    kt : float
        Morphologic age of template in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units

    Methods
    -------
    template():
        Returns array of windowed template function
    template_numexpr():
        Returns array of windowed template function optimized using numexpr

    References
    ----------

    Adapted from template derived in

    Hilley, G.E., DeLong, S., Prentice, C., Blisniuk, K. and Arrowsmith, 
    J.R., 2010. Morphologic dating of fault scarps using airborne 
    laser swath mapping (ALSM) data. Geophysical Research Letters, 37(4).
    https://dx.doi.org/10.1029/2009GL042044

    Based on solutions to the diffusion equation published in

    Hanks, T.C., 2000. The age of scarplike landforms from 
    diffusion‐equation analysis. Quaternary geochronology, 4, pp.313-338.

    and many references therein.
    """

    def __init__(self, d, kt, alpha, nx, ny, de):
        """Constructor method for scarp template

        Attributes
        ----------
        d : float
            Scale of windowed template function in data projection units
        kt : float
            Morphologic age of template in m2
        alpha : float
            Orientation of windowed template function in radians
        nx : int
            Number of columns in template array
        ny : int
            Number of rows in template array
        de : float
            Spacing of template grid cells in dat projection units
        """

        self.d = d
        self.kt = kt
        self.alpha = -alpha
        self.nx = nx
        self.ny = ny
        self.de = de

        frac = 0.9
        self.c = abs(2 * np.sqrt(self.kt) * erfinv(frac))

    def template(self):
        """Return template function

        Returns
        -------
        W : numpy array
            Windowed template function
        """

        x = self.de * np.linspace(1, self.nx, num=self.nx)
        y = self.de * np.linspace(1, self.ny, num=self.ny)
        x = x - np.mean(x)
        y = y - np.mean(y)

        x, y = np.meshgrid(x, y)
        xr = x * np.cos(self.alpha) + y * np.sin(self.alpha)
        yr = -x * np.sin(self.alpha) + y * np.cos(self.alpha)

        W = (-xr / (2. * self.kt ** (3 / 2.) * np.sqrt(np.pi))) \
            * np.exp(-xr ** 2. / (4. * self.kt))

        mask = self.get_mask()
        W = W * mask

        return W

    def template_numexpr(self):
        """Return template function (uses numexpr where possible)

        Returns
        -------
        W : numpy array
            Windowed template function
        """

        alpha = self.alpha
        kt = self.kt
        c = self.c
        d = self.d

        x = self.de * np.linspace(1, self.nx, num=self.nx)
        y = self.de * np.linspace(1, self.ny, num=self.ny)
        x = x - np.mean(x)
        y = y - np.mean(y)

        x, y = np.meshgrid(x, y)
        xr = numexpr.evaluate("x * cos(alpha) + y * sin(alpha)")
        yr = numexpr.evaluate("-x * sin(alpha) + y * cos(alpha)")

        pi = np.pi
        W = numexpr.evaluate("(-xr / (2 * kt ** (3/2) * sqrt(pi))) * \
                             exp(-xr ** 2 / (4 * kt))")

        mask = numexpr.evaluate("(abs(xr) < c) & (abs(yr) < d)")
        W = numexpr.evaluate("W * mask")

        return W


class RightFacingUpperBreakScarp(Scarp):
    """Template for upper slope break of vertical scarp (right-facting)

    Overrides template function to correct facign direction

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    kt : float
        Morphologic age of template in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units

    Methods
    -------
    get_error_mask():
        Return mask array that masks the lower slope break of scarp
    template():
        Returns array of windowed template function
    """

    def template(self):
        """Return template function (uses numexpr where possible)

        Returns
        -------
        W : numpy array
            Windowed template function
        """
        W = super().template_numexpr()
        return -W

    def get_err_mask(self):
        """Return mask array masking the lower half of scarp

        Returns
        -------
        mask : numpy array
            Mask array for lower half of scarp
        """
        xr, _ = self.get_coordinates()
        mask = xr <= 0
        return mask


class LeftFacingUpperBreakScarp(Scarp):
    """Template for upper slope break of vertical scarp (left-facting)

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    kt : float
        Morphologic age of template in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units

    Methods
    -------
    get_error_mask():
        Return mask array that masks the lower slope break of scarp
    """

    def get_err_mask(self):
        """Return mask array masking the lower half of scarp

        Returns
        -------
        mask : numpy array
            Mask array for lower hald of scarp
        """
        xr, _ = self.get_coordinates()
        mask = xr >= 0
        return mask


class ShiftedTemplateMixin(WindowedTemplate):
    """Mix-in for template that is offset from the window center

    Overrides template function to shift template

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    kt : float
        Morphologic age of template in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units
    dx : float
        X Offset of template center in data projection units
    dy : float
        Y Offset of template center data projection units

    Methods
    -------
    set_offset(dx, dy):
        Set offset attrivutes t odx and dy
    shift_template(W, dx, dy):
        Shift template array W by dx and dy
    template():
        Returns array of windowed template function
    """

    def __init__(self, *args, **kwargs):
        """Constructor for shifted template

        Parameters
        ----------
        dx : float
            X Offset of template center in data projection units
        dy : float
            Y Offset of template center data projection units
        """
        super().__init__(*args)
        self.set_offset(kwargs['dx'], kwargs['dy'])

    def set_offset(self, dx, dy):
        """Set offset values

        Parameters
        ----------
        dx : float
            X Offset of template center in data projection units
        dy : float
            Y Offset of template center data projection units
        """

        self.dx = dx
        self.dy = dy

    def shift_template(self, W, dx, dy):
        """Shift template

        Parameters
        ----------
        W : numpy array
            Windowed template function
        dx : float
            X Offset of template center in data projection units
        dy : float
            Y Offset of template center data projection units

        Returns
        -------
        W : numpy array
            Shifted windowed template function
        """

        ny, nx = W.shape

        if dx > 0:
            left = np.zeros((ny, dx))
            W = W[:, 0:-dx]
            W = np.hstack([left, W])
        else:
            dx = abs(dx)
            right = np.zeros((ny, dx))
            W = W[:, dx:]
            W = np.hstack([W, right])

        if dy > 0:
            bottom = np.zeros((dy, nx))
            W = W[0:-dy, :]
            W = np.vstack([W, bottom])
        else:
            dy = abs(dy)
            top = np.zeros((dy, nx))
            W = W[dy:, :]
            W = np.vstack([top, W])

        return W

    def template(self):
        """Template function for shifted template

        Returns
        -------
        W : numpy array
            Shifted windowed template function
        """

        W = super().template()
        W = self.shift_template(W, self.dx, self.dy)
        return W


class ShiftedLeftFacingUpperBreakScarp(ShiftedTemplateMixin,
                                       LeftFacingUpperBreakScarp):
    pass


class ShiftedRightFacingUpperBreakScarp(ShiftedTemplateMixin,
                                        RightFacingUpperBreakScarp):
    pass


class Ricker(WindowedTemplate):
    """Template using 2D Ricker wavelet

    Attributes
    ----------
    d : float
        Scale of windowed template function in data projection units
    alpha : float
        Orientation of windowed template function in radians
    kt : float
        Morphologic age of template in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units

    Methods
    -------
    template():
        Returns array of windowed template function

    References
    ----------

    This implements a Ricker wavelet similar to thet used in the following work

    Lashermes, B., Foufoula‐Georgiou, E., and Dietrich, W. E., 2007, Channel 
    network extraction from high resolution topography using wavelets. 
    Geophysical Research Letters, 34(23).
    https://doi.org/10.1029/2007GL031140
    """

    def __init__(self, d, f, alpha, nx, ny, de):
        """Constructor method for Ricker template

        Paramters
        ---------
        d : float
            Scale of windowed template function in data projection units
        kt : float
            Morphologic age of template in m2
        alpha : float
            Orientation of windowed template function in radians
        nx : int
            Number of columns in template array
        ny : int
            Number of rows in template array
        de : float
            Spacing of template grid cells in dat projection units
        """

        self.d = d
        self.f = f
        self.alpha = -alpha
        self.nx = nx
        self.ny = ny
        self.c = nx
        self.de = de

    def get_window_limits(self):
        return np.zeros((self.ny, self.nx), dtype=bool)

    def template(self):
        """Template function for windowed Ricker wavelet 

        Returns
        -------
        W : numpy array
            Windowed template function
        """

        alpha = self.alpha
        d = self.d
        f = self.f

        xr, yr = self.get_coordinates() 

        pi = np.pi
        W = numexpr.evaluate("(1. - 2. * (pi * f * xr) ** 2.) * \
                             exp(-(pi * f * xr) ** 2.)")

        mask = self.get_mask()
        W = W * mask

        return W


class Channel(Ricker):
    """Duplicate class for Ricker wavelet used for fluvial channels"""
    pass


class Crater(WindowedTemplate):
    """Template for radially symmetric crater

    Attributes
    ----------
    r : float
        Radius of crater in pixels
    kt : float
        Morphologic age of template crater rim in m2
    nx : int
        Number of columns in template array
    ny : int
        Number of rows in template array
    de : float
        Spacing of template grid cells in dat projection units
    """

    def __init__(self, r, kt, nx, ny, de):
        """Constructor methodfor radially symmetric crater

        Parameters
        ----------
        r : float
            Radius of crater in pixels
        kt : float
            Morphologic age of template crater rim in m2
        nx : int
            Number of columns in template array
        ny : int
            Number of rows in template array
        de : float
            Spacing of template grid cells in dat projection units
        """

        self.r = r / de
        self.kt = kt
        self.nx = nx
        self.ny = ny
        self.de = de

    def template(self):
        """Template function for radially symmetric crater

        Returns
        -------
        W : numpy array
            Windowed template function
        """

        x = self.de * np.linspace(1, self.nx, num=self.nx)
        y = self.de * np.linspace(1, self.ny, num=self.ny)
        x = x - np.mean(x)
        y = y - np.mean(y)

        x, y = np.meshgrid(x, y)

        W = np.zeros_like(x)

        thetas = np.linspace(0, 2 * np.pi, num=359, endpoint=False)
        for theta in thetas:
            alpha = -theta
            dx = self.r * np.cos(theta)
            dy = self.r * np.sin(theta)
            xr = (x - dx) * np.cos(alpha) + (y + dy) * np.sin(alpha)
            yr = -(x - dx) * np.sin(alpha) + (y + dy) * np.cos(alpha)
            this_W = (-xr / (2. * self.kt ** (3 / 2.) * np.sqrt(np.pi))) \
                    * np.exp(-xr ** 2. / (4. * self.kt))

            mask = (abs(xr) < 1) & (abs(yr) < 5 / self.de)

            this_W *= mask

            if theta > np.pi / 2 and theta < 3 * np.pi / 2:
                this_W *= -1

            W += this_W

        return W



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scarplet",
    version="0.1.1",
    author=["Robert Sare", "George Hilley"],
    author_email="rmsare@stanford.edu",
    description="A Python package for topographic template matching",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/rmsare/scarplet",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_require=">=3.4",
    install_requires=[
        "numexpr",
        "numpy",
        "matplotlib",
        "gdal",
        "pyfftw",
        "rasterio",
        "scipy"
    ]
)
